def go():
    """ Application entry point """
    import logging

    from histgrinder.config import read_configuration, lookup_name
    from histgrinder.transform import Transformer

    # set up arguments
    from argparse import ArgumentParser

    parser = ArgumentParser(description="Histogram postprocessing script")
    parser.add_argument('-c', '--configfile', nargs='+',
                        help='YAML configuration file(s)')
    parser.add_argument('source', help='Input source, e.g. ROOT file')
    parser.add_argument('target', help='Output target, e.g. ROOT file')
    parser.add_argument('--inmodule',
                        default='histgrinder.io.root.ROOTInputModule',
                        help='Python class which implements an input module')
    parser.add_argument('--outmodule',
                        default='histgrinder.io.root.ROOTOutputModule',
                        help='Python class which implements an output module')
    parser.add_argument('--prefix',
                        help='Prefix to ignore in histogram locations')
    parser.add_argument('--loglevel', help='Set the logging level',
                        choices=['DEBUG', 'INFO', 'WARNING',
                                 'ERROR', 'CRITICAL'],
                        default='INFO')
    args = parser.parse_args()

    logging.basicConfig(level=args.loglevel,
                        format='%(asctime)s | %(name)s | %(levelname)s | %(message)s')  # noqa: E501
    log = logging.getLogger(__name__)
    log.info("Histgrinder: histogram postprocessor")

    # read configuration & set up transformations
    transformers = []
    for configfile in args.configfile:
        config = read_configuration(configfile)
        transformers += [Transformer(_) for _ in config]
    selectors = set()
    for transform in transformers:
        selectors.update(transform.inregexes)

    # Configure input
    im = lookup_name(args.inmodule)()
    in_configuration = {'source': args.source}
    if args.prefix:
        in_configuration['prefix'] = args.prefix
    im.configure(in_configuration)
    im.setSelectors(selectors)

    # Configure output
    om = lookup_name(args.outmodule)()
    out_configuration = {'target': args.target}
    if args.prefix:
        out_configuration['prefix'] = args.prefix
    om.configure(out_configuration)

    # Warmup
    log.info("Warmup")
    for obj in im.warmup():
        for _ in transformers:
            v = _.consider(obj)

    # Event loop
    log.info("Beginning loop")
    for obj in im:
        for _ in transformers:
            v = _.consider(obj)
            if v:
                om.publish(v)

    log.info("Complete")


if __name__ == '__main__':
    go()



from ..interfaces import InputModule, OutputModule
from ..HistObject import HistObject
from typing import (Union, Iterable, Mapping, Any, Collection,
                    Pattern, Generator)
import logging


def readobj(k, dryrun):
    if not dryrun:
        obj = k.ReadObj()
        if hasattr(obj, 'SetDirectory'):
            obj.SetDirectory(0)
    else:
        obj = None
    return obj


class ROOTInputModule(InputModule):
    def __init__(self):
        self.source = None
        import ROOT
        # allow for us to close the input file and keep the histograms live
        ROOT.TH1.AddDirectory(ROOT.kFALSE)
        self.classwarnings = set()
        self.selectors = None

    def configure(self, options: Mapping[str, Any]) -> None:
        """
        Configure this module. Potential elements of "options":
        source: should be a ROOT-openable filename or URL.
        prefix: directory path to search under. Returned histogram names
            will not include this.
        """
        if 'source' not in options:
            raise ValueError("Must specify 'source' as an "
                             "option to ROOTInputModule")
        self.source = options['source']
        self.prefix = options.get('prefix', '/')

    def setSelectors(self, selectors: Collection[Pattern]) -> None:
        """ Do more later """
        self.selectors = selectors

    def iterate(self, dryrun) -> Generator[HistObject, None, None]:
        """ Open ROOT file; iterate over all histograms; close ROOT file """
        import ROOT
        import os.path
        from collections import deque
        log = logging.getLogger(__name__)
        infile = ROOT.TFile.Open(self.source)
        if not infile:
            raise ValueError(f"Unable to open input file {self.source}")
        dirqueue = deque([''])
        while dirqueue:
            dirname = dirqueue.popleft()
            indir = infile.GetDirectory(os.path.join(self.prefix, dirname))
            if not indir:
                log.critical("Access to invalid directory. "
                             f"This shouldn't happen ... dirname {dirname}")
                continue
            for k in indir.GetListOfKeys():
                classname = k.GetClassName()
                if classname.startswith('TDirectory'):
                    dirqueue.append(os.path.join(dirname, k.GetName()))
                    continue
                klass = ROOT.TClass.GetClass(classname)
                if not (klass.InheritsFrom('TH1')
                        or klass.InheritsFrom('TGraph')
                        or klass.InheritsFrom('TEfficiency')):
                    self._classwarning(k, classname, log)
                    continue
                objname = os.path.join(dirname, k.GetName())
                if self.selectors is not None:
                    if not any(_.match(objname) for _ in self.selectors):
                        continue
                obj = readobj(k, dryrun)
                log.debug('ROOT input read '
                          f'{os.path.join(dirname, k.GetName())}')
                yield HistObject(os.path.join(dirname, k.GetName()), obj)
        infile.Close()

    def _classwarning(self, key, classname, log) -> None:
        """ Log warning for unhandled class """
        if classname not in self.classwarnings:
            self.classwarnings.add(classname)
            log.warning(f"{key.GetName()} is of type {classname} "
                        "and cannot be considered, for now")
            log.warning(f"Future warnings for type {classname} "
                        "will be suppressed")

    def __iter__(self) -> Iterable[HistObject]:
        return self.iterate(dryrun=False)

    def warmup(self) -> Iterable[HistObject]:
        return self.iterate(dryrun=True)


class ROOTOutputModule(OutputModule):
    def __init__(self):
        self.target = None

    def configure(self, options: Mapping[str, Any]) -> None:
        """
        Configure this module. Potential elements of "options":
        target: should be a ROOT-openable filename or URL which
            can be opened for writing.
        prefix: directory path to place results under.
        overwrite: boolean to indicate whether results should overwrite
            existing histograms in the file.
        delay: only write histograms in finalize() (not during publish()).
        """
        if 'target' not in options:
            raise ValueError("Must specify 'target' as an option "
                             "to ROOTInputModule")
        self.target = options['target']
        self.overwrite = bool(options.get('overwrite', True))
        self.prefix = options.get('prefix', '/')
        self.delay = bool(options.get('delay', False))
        self.queue = set()

    def publish(self, obj: Union[HistObject, Iterable[HistObject]]) -> None:
        """ Accepts a HistObject containing a ROOT object to write to file """
        if isinstance(obj, HistObject):
            obj = [obj]
        if self.delay:
            self.queue.update(obj)
        else:
            self.queue = obj
            self._write()
            self.queue = None

    def _write(self) -> None:
        """ Open ROOT file; write obj; close ROOT file """
        import ROOT
        import os.path
        if not self.queue:
            return  # Nothing to do
        log = logging.getLogger(__name__)
        outfile = ROOT.TFile.Open(self.target, 'UPDATE')
        for o in self.queue:
            log.debug(f"ROOT output: publishing {o}")
            fulltargetname = os.path.join(self.prefix, o.name)
            dirtargetname = os.path.dirname(fulltargetname)
            if isinstance(o.hist, ROOT.TObject):
                if not outfile.GetDirectory(dirtargetname):
                    outfile.mkdir(dirtargetname if dirtargetname[0] != '/'
                                  else dirtargetname[1:])
                d = outfile.GetDirectory(dirtargetname)
                d.WriteTObject(o.hist, os.path.basename(fulltargetname),
                               "WriteDelete" if self.overwrite else "")
            else:
                log.error("ROOT output: unsupported object type "
                          f"{type(o.hist).__name__}")
        outfile.Close()

    def finalize(self) -> None:
        """ Writes outstanding HistObjects to file """
        self._write()


if __name__ == '__main__':
    import sys
    if len(sys.argv) != 3:
        print("Provide two arguments (input and output files)")
        sys.exit(1)
    im = ROOTInputModule()
    im.configure({'source': sys.argv[1]})
    om = ROOTOutputModule()
    om.configure({'target': sys.argv[2]})
    for o in im:
        print(o, o.hist.GetMean())
        om.publish(o)



# collect example code


def transform_function_divide_ROOT(inputs):
    """ This function returns the ratio of two ROOT histograms """
    # there should only be one match at a time. inputs[0] will describe that match.
    assert len(inputs) == 1  # for testing
    # inputs[0] is a pair of (dict describing the regex matches, list of histograms)
    # so we expect inputs[0][1] to be a pair, should have length 2
    assert len(inputs[0][1]) == 2
    cl = inputs[0][1][0].Clone()
    cl.Divide(inputs[0][1][1])
    return [cl]


def transform_function_divide2_ROOT(inputs):
    """ This function returns the ratio of two ROOT histograms """
    assert len(inputs) == 10
    cl = inputs[0][1][0].Clone()
    cl.Clear()
    for o in inputs:
        cl2 = o[1][0].Clone()
        cl2.Divide(o[1][1])
        cl.Add(cl2)
    return [cl]


def transform_function_rms_ROOT(inputs):
    """ This function returns the ratio of two ROOT histograms """
    import ROOT
    rv = ROOT.TH1F('RMS', 'RMS of gaussians', 40, 0, 2)
    plots = [_[1][0] for _ in inputs]  # all plots passed as first element of list
    for plot in plots:
        rv.Fill(plot.GetRMS())
    return [rv]


def nop(inputs):
    """ This function does nothing but may be useful for debugging """
    print(list(inputs))
    return []



from typing import Any


# Object to hold histogram name & object
class HistObject(object):
    def __init__(self, name: str, hist: Any):
        self.name = name
        self.hist = hist

    def __str__(self):
        return f"{type(self.hist).__name__} object {self.name}"



from .config import TransformationConfiguration, lookup_name
from .HistObject import HistObject
from typing import Optional, List


class Transformer(object):
    def __init__(self, tc: TransformationConfiguration):
        import re
        import string
        self.tc = tc
        # the number of histograms needed for a match
        self.inlength = len(tc.input)
        # regexes of input
        self.inregexes = [re.compile(_) for _ in tc.input]
        # info for pattern matching
        self.regexgroups = [_.groupindex for _ in self.inregexes]
        # check that all regex group keys are subsets of the first pattern
        for regexgroup in self.regexgroups[1:]:
            if not regexgroup.keys() <= self.regexgroups[0].keys():
                raise ValueError("Pattern specification problem: all named "
                                 "groups must be given in the first pattern")
        self.regextupnames = [tuple(_) for _ in self.regexgroups]

        # group names that appear in outputs
        self.outputnames = set().union(*[[_[1] for _ in string.Formatter().parse(output)]
                                         for output in self.tc.output])

        # one dictionary for each input slot
        self.hits = [{} for _ in range(len(self.inregexes))]
        try:
            self.transform_function = lookup_name(tc.function)
            if not callable(self.transform_function):
                raise ValueError(f"{tc.function} does not appear "
                                 "to be callable")
        except Exception as e:
            raise ValueError(f"Unable to instantiate transformer because: {e}")

    def consider(self, obj: HistObject, dryrun: bool = False) -> Optional[List[HistObject]]:
        """ Emit a new plot if we get a full match, otherwise None """
        import logging
        log = logging.getLogger(__name__)
        log.debug(self.tc.description)
        match = None
        for ire, regex in enumerate(self.inregexes):
            imatch = regex.match(obj.name)
            if imatch:
                self.hits[ire][imatch.groups()] = obj
                match = imatch
        if match is None:
            return None
        # Return value list
        rv = []

        # Given a match, what first position histograms are relevant?
        firstmatches = self._getMatchingFirstHists(match)

        # Group the first position matches by integration variables
        groupedfirstmatches = self._groupMatches(firstmatches)

        # construct iterables & call functions
        for tuplist in groupedfirstmatches.values():
            hci = HistCombinationIterable(self, tuplist)
            if _fullyvalid(hci):
                olist = self.transform_function(hci, **self.tc.parameters)
                for i, ohist in enumerate(olist):
                    oname = self.tc.output[i].format(**dict(zip(self.regextupnames[0], tuplist[0])))
                    rv.append(HistObject(oname, ohist))

        return rv

    def _getMatchingFirstHists(self, match):
        firstmatches = []
        for tup in self.hits[0]:
            # does the tuple match in all spots where the pattern name matches, and
            # where the variable is significant (we would output a different plot)?
            # make sure we handle the 1-indexing of regex matches
            if all(tup[self.regexgroups[0][_[0]]-1] == _[1]
                   for _ in match.groupdict().items()
                   if _[0] in self.outputnames):
                firstmatches.append(tup)
        return firstmatches

    def _groupMatches(self, firstmatches):
        # group matches by self.outputnames
        import collections
        rv = collections.defaultdict(list)
        for tup in firstmatches:
            reducedtup = tuple(v for i, v in enumerate(tup)
                               if self.regextupnames[0][i] in self.outputnames)
            rv[reducedtup].append(tup)
        return rv


class HistCombinationIterable(object):
    def __init__(self, transformer: Transformer, tuples):
        self.transform = transformer
        # tuple of matching first-position objects.
        # corresponds to each iteration step.
        self.tuples = tuples

    def __iter__(self):
        # logic: every tuple (first position object) should correspond to
        # exactly one histogram in each other slot. Find the full set and
        # yield as a tuple ({matches}, [histograms])
        for tup in self.tuples:
            rv = self._getpair(tup)
            if rv is None:
                continue
            yield rv

    def _getpair(self, tup):
        t = self.transform
        matchdict = dict(zip(t.regextupnames[0], tup))
        matchlist = [[tup]]
        for idx in range(1, t.inlength):
            matchlist.append([])
            for tup2 in t.hits[idx]:
                if all(v == tup[t.regexgroups[0][t.regextupnames[idx][i]]-1]  # noqa: E501
                        for i, v in enumerate(tup2)):
                    matchlist[-1].append(tup2)
        if any(not _ for _ in matchlist):
            return None
        return (matchdict, [t.hits[i][v[0]].hist for i, v in enumerate(matchlist)])

    def __getitem__(self, idx):
        return self._getpair(self.tuples[idx])

    def __len__(self):
        return len(self.tuples)


def _fullyvalid(hci: HistCombinationIterable) -> bool:
    hasany = False  # return false if there are no valid combinations
    for o in hci:
        hasany = True
        if any(_ is None for _ in o[1]):
            return False
    return hasany



if __name__ == '__main__':
    import ROOT
    import random
    random.seed(42)

    f = ROOT.TFile.Open('example.root', 'RECREATE')
    f.mkdir('prefix/gaussians')
    f.cd('prefix/gaussians')
    for i in range(100):
        h = ROOT.TH1F(f'gaus_{i}', f'Gaussian {i}', 40, -4, 4)
        for j in range(1000):
            h.Fill(random.gauss(0, 1))
        h.Write()
    f.Close()



# Configuration utilities
from typing import Union, IO, List, Any
import logging


class TransformationConfiguration(object):
    def __init__(self, Input, Output, Function, Description, Parameters={}):
        self.input = Input
        self.output = Output
        self.function = Function
        self.parameters = Parameters
        self.description = Description

    def __repr__(self):
        return (f"Description: {self.description}\nInput: {self.input}\n"
                f"Output: {self.output}\n"
                f"Function: {self.function}, Parameters: {self.parameters}")


def read_configuration(f: Union[str, IO]) -> List[TransformationConfiguration]:
    import yaml
    if isinstance(f, str):
        fobj = open(f)
    else:
        fobj = f

    rv = []
    for doc in yaml.safe_load_all(fobj):
        try:
            rv.append(TransformationConfiguration(**doc))
        except TypeError as e:
            log = logging.getLogger(__name__)
            log.error("Unable to understand the following configuration:")
            log.error(doc)
            log.error(f"Error message: {e}")
    return rv


def lookup_name(name: str) -> Any:
    import importlib
    name = name.rsplit('.', 1)
    return getattr(importlib.import_module(name[0]), name[1])


if __name__ == "__main__":
    import sys
    for _ in read_configuration(sys.argv[1]):
        print(_)



from abc import ABC, abstractmethod
from typing import Mapping, Collection, Iterable, Union
from .HistObject import HistObject


# Interface for modules that read histograms from a source
class InputModule(ABC):
    @abstractmethod
    def configure(self, options: Mapping) -> None:
        """
        Called to configure the module. Definition is up to specific
        implementation.
        """
        return

    @abstractmethod
    def setSelectors(self, selectors: Collection[str]) -> None:
        return

    @abstractmethod
    def __iter__(self) -> Iterable[HistObject]:
        return None

    @abstractmethod
    def warmup(self) -> Iterable[HistObject]:
        """
        Should return an iterable that behaves the same as normal iteration,
        except with HistObject.hist = None (histograms not actually read).
        This permits code to figure out the structure of the histogram file.
        """
        return None


# Interface for modules that write histograms to a sink
class OutputModule(ABC):
    @abstractmethod
    def configure(self, options: Mapping) -> None:
        """ Called to configure the module."""
        return

    @abstractmethod
    def publish(self, obj: Union[HistObject, Iterable[HistObject]]) -> None:
        """ Publish the histogram. """
        return

    @abstractmethod
    def finalize(self) -> None:
        return



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="histgrinder",
    version="0.1.3",
    author="Peter Onyisi",
    author_email="ponyisi@utexas.edu",
    description="Generic system to perform streaming transformations on histograms",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ponyisi/histogram_postprocessing",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Physics",
    ],
    python_requires='>=3.6',
    install_requires=['PyYAML>=5'],
    scripts=['bin/histgrinder']
)
